//! Relayers
//!
//! This is the main entry point to the client library.
use crate::{Client, Error};
use futures::future::join_all;
use nostr_rs_types::{
    client::{self, subscribe},
    Response,
};
use std::{
    collections::HashMap,
    sync::{
        atomic::{AtomicUsize, Ordering},
        Arc,
    },
    time::Duration,
};
use subscription::Scheduler;
use tokio::{
    sync::{mpsc, RwLock},
    time::sleep,
};
use url::Url;

pub mod subscription;

pub(crate) type AllClients = Arc<RwLock<HashMap<Url, (Arc<AtomicUsize>, Client)>>>;

/// Clients
///
/// This is a set of outgoing connections to relayers. This struct can connect
/// async to N relayers offering a simple API to talk to all of them at the same
/// time, and to receive messages
pub struct Pool {
    clients: AllClients,
    sender: mpsc::Sender<(Response, Url)>,
    receiver: Option<mpsc::Receiver<(Response, Url)>>,
    subscription_manager: Arc<subscription::Scheduler>,
}

/// Active client
///
/// For each connection on the pool this object will be returned. When dropped,
/// that connection is also dropped from the connection pool.
pub struct ActiveClient {
    client_id: Url,
    counter: Arc<AtomicUsize>,
    all_clients: AllClients,
}

impl Drop for ActiveClient {
    fn drop(&mut self) {
        let counter = self.counter.fetch_sub(1, Ordering::SeqCst);
        if counter == 0 {
            let all_clients = self.all_clients.clone();
            let client_id = self.client_id.clone();
            tokio::spawn(async move {
                // remove the client from the pool, when it goes out of scope
                // it will be disconnected
                all_clients.write().await.remove(&client_id);
            });
        }
    }
}

/// Default channel buffer size for the pool
pub const DEFAULT_CHANNEL_BUFFER_SIZE: usize = 100_000;

impl Default for Pool {
    fn default() -> Self {
        let (sender, receiver) = mpsc::channel(DEFAULT_CHANNEL_BUFFER_SIZE);
        let clients: AllClients = Default::default();
        Self {
            clients: clients.clone(),
            receiver: Some(receiver),
            subscription_manager: Arc::new(Scheduler::new(clients.clone())),
            sender,
        }
    }
}

impl Pool {
    /// Creates a new instance with a list of urls
    pub fn new_with_clients(clients: Vec<Url>) -> Result<(Self, Vec<ActiveClient>), Error> {
        let pool = Self::default();
        let connect = clients.into_iter().map(|url| pool.connect_to(url));

        let x = pool.subscription_manager.clone();
        tokio::spawn(async move {
            loop {
                log::info!("Active subscribers: {}", x.debug().await);
                sleep(Duration::from_secs(5)).await;
            }
        });

        futures::executor::block_on(async {
            futures::future::join_all(connect)
                .await
                .into_iter()
                .collect::<Result<Vec<_>, _>>()
        })
        .map(|clients| (pool, clients))
    }

    /// Splits the pool removing the receiver to be used in a different context
    pub fn split(mut self) -> Result<(mpsc::Receiver<(Response, Url)>, Self), Error> {
        Ok((self.receiver.take().ok_or(Error::AlreadySplitted)?, self))
    }

    /// Tries to receive a message from any of the connected relayers
    pub fn try_recv(&mut self) -> Option<(Response, Url)> {
        self.receiver.as_mut()?.try_recv().ok()
    }

    /// Receives a message from any of the connected relayers
    pub async fn recv(&mut self) -> Option<(Response, Url)> {
        self.receiver.as_mut()?.recv().await
    }

    /// Returns the number of active subscriptions
    pub async fn active_subscriptions(&self) -> usize {
        self.subscription_manager.total_subscribers()
    }

    /// Subscribes to all the connected relayers
    pub async fn subscribe(
        &self,
        subscription: subscribe::Subscribe,
    ) -> Result<subscription::ActiveSubscription, Error> {
        self.subscription_manager.subscribe(subscription).await
    }

    /// Sends a request to all the connected relayers
    pub async fn post(&self, request: client::Event) {
        let clients = self.clients.read().await;
        join_all(
            clients
                .values()
                .map(|(_, sender)| sender.post(request.clone()))
                .collect::<Vec<_>>(),
        )
        .await;
    }

    /// Returns the number of active connections.
    pub async fn check_active_connections(&self) -> usize {
        self.clients
            .read()
            .await
            .iter()
            .filter(|(_, (_, client))| client.is_connected())
            .collect::<Vec<_>>()
            .len()
    }

    /// Creates a connection to a new relayer.
    ///
    /// This function will open a connection at most once, if a connection
    /// already exists false will be returned
    pub async fn connect_to(&self, url: Url) -> Result<ActiveClient, Error> {
        let mut clients = self.clients.write().await;

        let ref_id = if let Some((id, _)) = clients.get(&url) {
            id.fetch_add(1, Ordering::SeqCst);
            id.clone()
        } else {
            log::warn!("Connecting to {}", url);
            let subscription_manager = self.subscription_manager.clone();
            let client = Client::new(
                self.sender.clone(),
                url.clone(),
                move |response, url, return_to| {
                    let subscription_manager = subscription_manager.clone();
                    Box::pin(async move {
                        subscription_manager
                            .process_message(response, url, return_to)
                            .await
                    })
                },
            );

            self.subscription_manager.register_client(&client).await;

            let ref_id: Arc<AtomicUsize> = Arc::new(1.into());
            clients.insert(url.clone(), (ref_id.clone(), client));
            ref_id
        };

        Ok(ActiveClient {
            client_id: url,
            counter: ref_id,
            all_clients: self.clients.clone(),
        })
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use nostr_rs_memory::Memory;
    use nostr_rs_relayer::Relayer;
    use nostr_rs_types::{
        account::Account,
        types::{Content, Filter},
    };
    use std::time::Duration;
    use subscription::MAX_ACTIVE_SUBSCRIPTIONS;
    use tokio::{net::TcpListener, task::JoinHandle, time::sleep};

    async fn dummy_server(port: u16) -> (Url, Arc<Relayer<Memory>>, JoinHandle<()>) {
        let listener = TcpListener::bind(format!("127.0.0.1:{}", port))
            .await
            .unwrap();
        let local_addr = listener.local_addr().expect("addr");

        let relayer = Relayer::new(Some(Memory::default()), None).expect("valid dummy server");
        let (relayer, stopper) = relayer.main(listener).expect("valid main loop");
        (
            Url::parse(&format!("ws://{}", local_addr)).expect("valid url"),
            relayer,
            stopper,
        )
    }

    #[tokio::test]
    async fn droppable_subscription() {
        let client_pool = Pool::default();
        let subscription = client_pool
            .subscribe(Default::default())
            .await
            .expect("valid");

        assert_eq!(client_pool.active_subscriptions().await, 1);
        drop(subscription);
        sleep(Duration::from_millis(10)).await;
        assert_eq!(client_pool.active_subscriptions().await, 0);
    }

    #[tokio::test]
    async fn connect_to_dummy_server() {
        let (addr, _, stopper) = dummy_server(0).await;
        let (client_pool, _connections) = Pool::new_with_clients(vec![addr]).expect("valid pool");

        assert_eq!(0, client_pool.check_active_connections().await);

        sleep(Duration::from_millis(10)).await;
        assert_eq!(1, client_pool.check_active_connections().await);

        // stop dummy server
        stopper.abort();

        sleep(Duration::from_millis(100)).await;
        assert_eq!(0, client_pool.check_active_connections().await);
    }

    #[tokio::test]
    async fn two_clients_communication() {
        let (addr, _, _) = dummy_server(0).await;
        let (mut client_pool1, _c1) =
            Pool::new_with_clients(vec![addr.clone()]).expect("valid pool");
        let (client_pool2, _c2) = Pool::new_with_clients(vec![addr]).expect("valid pool");

        let _sub1 = client_pool1.subscribe(Default::default()).await;

        sleep(Duration::from_millis(10)).await;

        assert!(client_pool1
            .try_recv()
            .map(|(r, _)| r)
            .expect("valid message")
            .as_end_of_stored_events()
            .is_some());
        assert!(client_pool1.try_recv().is_none());

        let account1 = Account::default();
        let signed_content = account1
            .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
            .expect("valid signed content");

        client_pool2.post(signed_content.clone().into()).await;

        sleep(Duration::from_millis(10)).await;

        assert_eq!(
            Some((signed_content.id, signed_content.signature)),
            client_pool1
                .try_recv()
                .and_then(|(r, _)| r.as_event().cloned().map(|x| x.event))
                .map(|x| (x.id, x.signature))
        );
        assert!(client_pool1.try_recv().is_none());
    }

    #[tokio::test]
    async fn reconnect_and_resubscribe() {
        let (addr, _, stopper) = dummy_server(0).await;
        let (mut client_pool1, _c1) =
            Pool::new_with_clients(vec![addr.clone()]).expect("valid pool");
        let (client_pool2, _c2) = Pool::new_with_clients(vec![addr.clone()]).expect("valid pool");

        let _sub1 = client_pool1.subscribe(Default::default()).await;

        sleep(Duration::from_millis(10)).await;

        assert!(client_pool1
            .try_recv()
            .map(|(r, _)| r)
            .expect("valid message")
            .as_end_of_stored_events()
            .is_some());
        assert!(client_pool1.try_recv().is_none());

        let account1 = Account::default();
        let signed_content = account1
            .sign_content(vec![], Content::ShortTextNote("test 1".to_owned()), None)
            .expect("valid signed content");

        client_pool2.post(signed_content.clone().into()).await;

        sleep(Duration::from_millis(10)).await;

        assert_eq!(
            Some((signed_content.id, signed_content.signature)),
            client_pool1
                .try_recv()
                .and_then(|(r, _)| r.as_event().cloned().map(|x| x.event))
                .map(|x| (x.id, x.signature))
        );
        assert!(client_pool1.try_recv().is_none());

        assert_eq!(1, client_pool1.check_active_connections().await);
        assert_eq!(1, client_pool2.check_active_connections().await);

        stopper.abort();

        sleep(Duration::from_millis(10)).await;

        assert_eq!(0, client_pool1.check_active_connections().await);
        assert_eq!(0, client_pool2.check_active_connections().await);

        let (_, _, stopper) = dummy_server(addr.port().expect("port")).await;

        sleep(Duration::from_millis(2_000)).await;

        assert_eq!(1, client_pool1.check_active_connections().await);
        assert_eq!(1, client_pool2.check_active_connections().await);

        let signed_content = account1
            .sign_content(vec![], Content::ShortTextNote("test 1".to_owned()), None)
            .expect("valid signed content");

        client_pool2.post(signed_content.clone().into()).await;

        sleep(Duration::from_millis(10)).await;

        assert_eq!(
            Some((signed_content.id, signed_content.signature)),
            client_pool1
                .try_recv()
                .and_then(|(r, _)| r.as_event().cloned().map(|x| x.event))
                .map(|x| (x.id, x.signature))
        );
        assert!(client_pool1.try_recv().is_none());

        stopper.abort();
    }

    #[tokio::test]
    async fn connect_multiple_servers() {
        let (addr1, _, _) = dummy_server(0).await;
        let (addr2, _, _) = dummy_server(0).await;
        let (mut client_pool1, _c1) =
            Pool::new_with_clients(vec![addr1.clone(), addr2]).expect("valid pool");
        let (client_pool2, _c2) = Pool::new_with_clients(vec![addr1]).expect("valid pool");

        let _sub1 = client_pool1.subscribe(Default::default()).await;

        sleep(Duration::from_millis(10)).await;

        assert!(client_pool1
            .try_recv()
            .map(|(r, _)| r)
            .expect("valid message")
            .as_end_of_stored_events()
            .is_some());

        let account1 = Account::default();
        let signed_content = account1
            .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
            .expect("valid signed content");

        client_pool2.post(signed_content.clone().into()).await;

        sleep(Duration::from_millis(10)).await;

        assert_eq!(
            Some((signed_content.id, signed_content.signature)),
            client_pool1
                .try_recv()
                .and_then(|(r, _)| r.as_event().cloned().map(|x| x.event))
                .map(|x| (x.id, x.signature))
        );
        assert!(client_pool1.try_recv().is_none());
    }

    /// Client pool user creates 101 subscription (that is not allowed by many
    /// relays), so the pool will do a round-robin subscriptions keeping 5
    /// actives at a time with an internal scheduler.
    ///
    /// The scheduler will pause active subscriptions when the EOS is received.
    /// On the next round `since` will be used to receive only newer events
    mod scheduler {
        use super::*;

        #[tokio::test]
        async fn stored_first() {
            let (addr1, relayer, _) = dummy_server(0).await;

            let (mut client_pool1, _c1) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");
            let (client_pool2, _c2) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");

            let account1 = Account::default();
            let signed_content = account1
                .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
                .expect("valid signed content");
            client_pool2.post(signed_content.clone().into()).await;

            let first_sub = client_pool1
                .subscribe(
                    Filter {
                        ids: vec![signed_content.id.clone()],
                        ..Default::default()
                    }
                    .into(),
                )
                .await
                .expect("valid subs, last");

            let subs = join_all(
                (0..100)
                    .into_iter()
                    .map(|_| {
                        client_pool1.subscribe(
                            Filter {
                                authors: vec![
                                "npub1k2q4dqk0eqlu6tp6m5zhsh852u7a8zz9wp5ewnxxmrx2q6eu8duq3ydzzr"
                                    .parse()
                                    .unwrap(),
                            ],
                                ..Default::default()
                            }
                            .into(),
                        )
                    })
                    .collect::<Vec<_>>(),
            )
            .await
            .into_iter()
            .collect::<Result<Vec<_>, _>>()
            .expect("valid 100 dummy subs");

            for _ in 1..10 {
                sleep(Duration::from_millis(10)).await;
                assert!(
                    MAX_ACTIVE_SUBSCRIPTIONS * 2 > relayer.total_subscribers(),
                    "total subs {}",
                    relayer.total_subscribers()
                );
            }

            sleep(Duration::from_secs(1)).await;

            let mut has_receive_event = false;
            for _ in 0..102 {
                let event = client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid event");

                if has_receive_event {
                    assert!(event.as_end_of_stored_events().is_some());
                } else {
                    if let Some(ev) = event.as_event() {
                        assert_eq!(ev.id, signed_content.id);
                        has_receive_event = true;
                    } else {
                        assert!(event.as_end_of_stored_events().is_some());
                    }
                }
            }

            drop(subs);
            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                1,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );

            drop(first_sub);
            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                0,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );
        }

        #[tokio::test]
        async fn stored_last() {
            let (addr1, relayer, _) = dummy_server(0).await;

            let (mut client_pool1, _c1) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");
            let (client_pool2, _c2) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");

            let account1 = Account::default();
            let signed_content = account1
                .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
                .expect("valid signed content");
            client_pool2.post(signed_content.clone().into()).await;

            let mut subs = join_all(
                (0..100)
                    .into_iter()
                    .map(|_| {
                        client_pool1.subscribe(
                            Filter {
                                authors: vec![
                                "npub1k2q4dqk0eqlu6tp6m5zhsh852u7a8zz9wp5ewnxxmrx2q6eu8duq3ydzzr"
                                    .parse()
                                    .unwrap(),
                            ],
                                ..Default::default()
                            }
                            .into(),
                        )
                    })
                    .collect::<Vec<_>>(),
            )
            .await
            .into_iter()
            .collect::<Result<Vec<_>, _>>()
            .expect("valid 100 dummy subs");

            for _ in 1..10 {
                sleep(Duration::from_millis(10)).await;
                assert!(
                    MAX_ACTIVE_SUBSCRIPTIONS * 2 > relayer.total_subscribers(),
                    "total subs {}",
                    relayer.total_subscribers()
                );
            }

            for _ in 0..100 {
                assert!(client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_end_of_stored_events()
                    .is_some());
            }

            assert!(client_pool1.try_recv().is_none());

            subs.push(
                client_pool1
                    .subscribe(
                        Filter {
                            ids: vec![signed_content.id.clone()],
                            ..Default::default()
                        }
                        .into(),
                    )
                    .await
                    .expect("valid subs, last"),
            );

            sleep(Duration::from_secs(1)).await;

            assert_eq!(
                client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_event()
                    .expect("valid event")
                    .id,
                signed_content.id
            );

            drop(subs);

            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                0,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );
        }

        #[tokio::test]
        async fn realtime_first() {
            let (addr1, relayer, _) = dummy_server(0).await;

            let (mut client_pool1, _c1) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");
            let (client_pool2, _c2) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");

            let account1 = Account::default();
            let signed_content = account1
                .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
                .expect("valid signed content");

            let first_sub = client_pool1
                .subscribe(
                    Filter {
                        ids: vec![signed_content.id.clone()],
                        ..Default::default()
                    }
                    .into(),
                )
                .await
                .expect("valid subs, first");

            let subs = join_all(
                (0..100)
                    .into_iter()
                    .map(|_| {
                        client_pool1.subscribe(
                            Filter {
                                authors: vec![
                                "npub1k2q4dqk0eqlu6tp6m5zhsh852u7a8zz9wp5ewnxxmrx2q6eu8duq3ydzzr"
                                    .parse()
                                    .unwrap(),
                            ],
                                ..Default::default()
                            }
                            .into(),
                        )
                    })
                    .collect::<Vec<_>>(),
            )
            .await
            .into_iter()
            .collect::<Result<Vec<_>, _>>()
            .expect("valid 100 dummy subs");

            for _ in 1..10 {
                sleep(Duration::from_millis(10)).await;
                assert!(
                    MAX_ACTIVE_SUBSCRIPTIONS * 2 > relayer.total_subscribers(),
                    "total subs {}",
                    relayer.total_subscribers()
                );
            }

            for _ in 0..101 {
                assert!(client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_end_of_stored_events()
                    .is_some());
            }

            assert!(client_pool1.try_recv().is_none());

            client_pool2.post(signed_content.clone().into()).await;

            sleep(Duration::from_secs(1)).await;

            assert_eq!(
                client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_event()
                    .expect("valid event")
                    .id,
                signed_content.id
            );

            drop(subs);

            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                1,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );

            drop(first_sub);

            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                0,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );
        }

        #[tokio::test]
        async fn realtime_last() {
            let (addr1, relayer, _) = dummy_server(0).await;

            let (mut client_pool1, _c1) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");
            let (client_pool2, _c2) =
                Pool::new_with_clients(vec![addr1.clone()]).expect("valid pool");

            let account1 = Account::default();
            let signed_content = account1
                .sign_content(vec![], Content::ShortTextNote("test 0".to_owned()), None)
                .expect("valid signed content");

            let mut subs = join_all(
                (0..100)
                    .into_iter()
                    .map(|_| {
                        client_pool1.subscribe(
                            Filter {
                                authors: vec![
                                "npub1k2q4dqk0eqlu6tp6m5zhsh852u7a8zz9wp5ewnxxmrx2q6eu8duq3ydzzr"
                                    .parse()
                                    .unwrap(),
                            ],
                                ..Default::default()
                            }
                            .into(),
                        )
                    })
                    .collect::<Vec<_>>(),
            )
            .await
            .into_iter()
            .collect::<Result<Vec<_>, _>>()
            .expect("valid 100 dummy subs");

            for _ in 1..10 {
                sleep(Duration::from_millis(10)).await;
                assert!(
                    MAX_ACTIVE_SUBSCRIPTIONS * 2 > relayer.total_subscribers(),
                    "total subs {}",
                    relayer.total_subscribers()
                );
            }

            for _ in 0..100 {
                assert!(client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_end_of_stored_events()
                    .is_some());
            }

            assert!(client_pool1.try_recv().is_none());

            subs.push(
                client_pool1
                    .subscribe(
                        Filter {
                            ids: vec![signed_content.id.clone()],
                            ..Default::default()
                        }
                        .into(),
                    )
                    .await
                    .expect("valid subs, last"),
            );

            sleep(Duration::from_secs(1)).await;

            assert!(client_pool1
                .try_recv()
                .map(|(r, _)| r)
                .expect("valid message")
                .as_end_of_stored_events()
                .is_some());

            client_pool2.post(signed_content.clone().into()).await;

            sleep(Duration::from_secs(1)).await;

            assert_eq!(
                client_pool1
                    .try_recv()
                    .map(|(r, _)| r)
                    .expect("valid message")
                    .as_event()
                    .expect("valid event")
                    .id,
                signed_content.id
            );

            drop(subs);

            sleep(Duration::from_secs(1)).await;
            assert_eq!(
                0,
                relayer.total_subscribers(),
                "total subs {}",
                relayer.total_subscribers()
            );
        }
    }
}
