//! Local connection
//!
//! Add types for adding local connections
use crate::{connection::ConnectionId, Error, Relayer};
use nostr_rs_storage_base::Storage;
use nostr_rs_types::{Request, Response};
use std::{sync::Arc, time::Duration};
use tokio::{
    sync::mpsc::{Receiver, Sender},
    task::JoinHandle,
    time::sleep,
};

/// Local connection
pub struct LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    /// The connection ID
    pub conn_id: ConnectionId,
    sender: Sender<(ConnectionId, Request)>,
    receiver: Receiver<Response>,
    relayer: Arc<Relayer<T>>,
}

impl<T> LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    /// Number of queued messages
    pub fn queued_messages(&self) -> usize {
        self.receiver.len()
    }
    /// Receive a message from the relayer
    pub async fn recv(&mut self) -> Option<Response> {
        self.receiver.recv().await
    }

    /// Try to receive a message from the relayer
    pub fn try_recv(&mut self) -> Option<Response> {
        self.receiver.try_recv().ok()
    }

    /// Queues sending a message to the relayer in the future time
    pub fn future_send(&self, request: Request, in_the_future: Duration) -> JoinHandle<()> {
        let sender = self.sender.clone();
        let conn_id = self.conn_id;
        tokio::spawn(async move {
            sleep(in_the_future).await;
            let _ = sender.send((conn_id, request)).await;
        })
    }

    /// Sends a request to the relayer
    pub async fn send(&self, request: Request) -> Result<(), Error> {
        self.sender
            .send((self.conn_id, request))
            .await
            .map_err(|e| Error::LocalSendError(Box::new(e)))
    }

    /// Get the sender
    pub fn sender(&self) -> Sender<(ConnectionId, Request)> {
        self.sender.clone()
    }
}

impl<T> Drop for LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    fn drop(&mut self) {
        self.relayer.drop_connection(self);
    }
}

impl<T>
    From<(
        ConnectionId,
        Receiver<Response>,
        Sender<(ConnectionId, Request)>,
        Arc<Relayer<T>>,
    )> for LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    fn from(
        value: (
            ConnectionId,
            Receiver<Response>,
            Sender<(ConnectionId, Request)>,
            Arc<Relayer<T>>,
        ),
    ) -> Self {
        LocalConnection {
            conn_id: value.0,
            receiver: value.1,
            sender: value.2,
            relayer: value.3,
        }
    }
}
